"""Implement a wrapper for primitive and strings."""
from dataclasses import dataclass
from typing import Callable
from .typevar import T, S
import abc


@dataclass
class Primitive(metaclass=abc.ABCMeta):
    """Wrapper for primitive and strings."""

    @abc.abstractproperty
    def primitive(self) -> T:
        """Return the raw value."""

    def handle(self, handler: Callable[[T], S]) -> S:
        """Apply a function to :py:meth:`primitive`."""
        return handler(self.primitive)



"""Provide methods to trace execution."""
import logging
import functools
from typing import Callable, Tuple, Any
import datetime
from dataclasses import dataclass, field


@dataclass
class Invocation:
    """Record an invocation of a function."""

    invocation: Tuple[Callable, Tuple, dict]
    start_time: datetime.datetime

    @classmethod
    def create_from(cls, function: Callable, args: Tuple, kwargs: dict):
        """Record the time when a `function` is called.

        Returns
        -------
        trace: Trace

        """
        return Invocation((function, args, kwargs), datetime.datetime.now())

    def build_start_message(self) -> str:
        """Create a message that represents this invocation."""
        function = self.invocation[0]
        args = self.invocation[1]
        kwargs = self.invocation[2]
        return '%s(%s, %s)' % (function.__name__, args, kwargs)

    def build_end_message(self, return_value: Any) -> str:
        """Create a message that this invocation has ended."""
        start_message = self.build_start_message()
        end_message = '%s -> %s' % (start_message, return_value)
        elapsed_time = self._count_elapsed_time()
        return '[elapsed time: %s milliseconds] %s' \
            % (elapsed_time, end_message)

    def _count_elapsed_time(self) -> float:
        end_time = datetime.datetime.now()
        delta = end_time - self.start_time
        return round(delta.total_seconds() * 1000, 3)


@dataclass
class TraceLevel:
    """Represent logging levels for tracing executions."""

    start_level: int = field(default=logging.DEBUG)
    end_level: int = field(default=logging.DEBUG)

    def start(self, logger: logging.Logger, invocation: Invocation) -> None:
        """Record `trace` with `logger`."""
        message = invocation.build_start_message()
        logger.log(self.start_level, message)

    def end(self,
            logger: logging.Logger,
            invocation: Invocation,
            return_value: Any) -> None:
        """Record that the function has end."""
        message = invocation.build_end_message(return_value)
        logger.log(self.end_level, message)


class Tracer:
    """Record information about a program's execution."""

    def __init__(self,
                 logger: logging.Logger,
                 trace_level: TraceLevel):
        """Take a logger and a logging level."""
        self.logger = logger
        self.trace_level = trace_level

    def trace(self, function: Callable):
        """Provide a function decorator to trace the decorated function."""
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            invocation = Invocation.create_from(function, args, kwargs)
            self.trace_level.start(self.logger, invocation)
            result = function(*args, **kwargs)
            self.trace_level.end(self.logger, invocation, result)
            return result
        return wrapper

    @classmethod
    def create_tracer(
            cls,
            logger_name: str = 'trace',
            format_string: str
            = '%(levelname)s:%(asctime)s:%(name)s:%(message)s',
            level: int = logging.DEBUG,
            trace_level: TraceLevel
            = TraceLevel(logging.DEBUG, logging.DEBUG)):
        """Create a tracer named with `logger_name`.

        Returns
        -------
        tracer: Tracer

        """
        handler = logging.StreamHandler()
        handler.setLevel(level)
        handler.setFormatter(logging.Formatter(format_string))
        logger = logging.getLogger(logger_name)
        logger.setLevel(level)
        logger.addHandler(handler)
        return Tracer(logger, trace_level)



"""Implement a wrapper for `str`."""
import os
from dataclasses import dataclass
from typing import Sequence
import collections.abc as collections


@dataclass
class Text:
    """Represent a text.

    Attributes
    ----------
    text: str

    """

    text: str

    def concat(self, text, separator='. '):
        """Concatenation of itself and `text`.

        Parameters
        ----------
        text: Text
            a text to be concatenated.

        """
        if text.is_empty():
            return self
        return Text(f'{self.text}{separator}{text.text}')

    def is_empty(self) -> bool:
        """Return `True` iff empty."""
        return len(self.text) == 0

    def strip(self):
        """Return the Stripped text.

        Returns
        -------
        Text

        """
        text = self.text.strip()
        return Text(text)

    def empty():
        """Return an emtpy :py:class:`Text`."""
        return Text('')

    def end_with_linesep(self):
        """Append a line separator if `self` does not end with it.

        Return
        ------
        Text

        """
        if self.text.endswith(os.linesep):
            return self
        return Text(f'{self.text}{os.linesep}')

    def tokenize(self, tokenizer):
        """Tokenize `self` to Texts.

        Parameters
        ----------
        tokenizer: function
            Take a `str`, and tokenize it
            to a sequence of :py:class:`Text`.

        Returns
        -------
        Texts

        """
        return Texts(tokenizer(self.text))


@dataclass
class Texts(collections.Sequence):
    """A sequence of :py:class:`Text`."""

    texts: Sequence[Text]

    def __len__(self):
        """Return the size of :py:attr:`texts`."""
        return len(self.texts)

    def __getitem__(self, s):
        """Access a subset of: py: attr: `texts`."""
        return self.texts.__getitem__(s)



"""Configure the logging system for snippets."""
import logging


class LogConfiguration:
    """Logging configuration.

    Attributes
    -----------
    verbose : bool

    name : str

    """

    def __init__(self, verbose: bool, name: str):
        """Take a bool value."""
        self.verbose = verbose
        self.name = name

    def configure(self) -> None:
        """Update the logging level of :py:attr:`name` logger."""
        logging.basicConfig(
            format='%(levelname)s:%(asctime)s:%(name)s:%(message)s')
        logging.getLogger(self.name).setLevel(
            logging.DEBUG if self.verbose else logging.INFO)



"""Provide classes that represent first class collections."""
import abc
import contextlib
from typing import Callable, Sequence, Generator, Container
from .typevar import T, S


class FirstClassFileContextManagerProvider(metaclass=abc.ABCMeta):
    """Provide a context manager for the content of :py:attr:`filename`.

    Attributes
    ----------
    filename: str
        The path to a file which :py:meth:`transform` transforms
        each line to an item emitted by :py:meth:`__call__`.

    """

    def __init__(self, filename: str):
        """Take the path to a file to read.

        Parameters
        ----------
        filename: str

        """
        self.filename = filename

    @contextlib.contextmanager
    def __call__(self) -> Generator[T, None, None]:
        """Emit an item transformed t from a line of :py:attr:`filename`."""
        with open(self.filename) as stream:
            yield (self.transform(item.strip()) for item in stream)

    @abc.abstractmethod
    def transform(self, item: str) -> T:
        """Transform each line of :py:attr:`filename`.

        Transform each line of :py:attr:`filename` to an item
        emitted by :py:meth:`__call__`.

        Parameters
        ----------
        item: str
            A line of :py:attr:`filename`.

        """


class FirstClassSequence(metaclass=abc.ABCMeta):
    """
    """

    @abc.abstractproperty
    def sequence(self) -> Sequence[T]:
        """Return the items."""

    def __getitem__(self, s):
        """Access the specified items."""
        return self.sequence.__getitem__(s)

    def __len__(self):
        """Return the size of :py:meth:`sequence`."""
        return len(self.sequence)

    def filter_by_container(self, container: Container[T]) \
            -> Generator[T, None, None]:
        """Return generator that emits items that container contains."""
        return (item for item in self.sequence if item in container)

    def is_empty(self) -> bool:
        """Return `True` if :py:meth:`sequence` is empty."""
        return len(self.sequence) == 0

    def apply_function(self, function: Callable[[T], S]) \
            -> Generator[S, None, None]:
        """
        """
        return (function(item) for item in self.sequence)




"""Information of greentea."""
from setuptools import setup, find_packages


setup(name='greentea',
      version="3.1.0",
      description=(
          'A microframework for abstraction.'),
      python_requires='>=3.8.0',
      url='https://github.com/nryotaro/greentea',
      author='Nakamura, Ryotaro',
      author_email='nakamura.ryotaro.kzs@gmail.com',
      license='MIT License',
      classifiers=['Programming Language :: Python :: 3.8'],
      packages=find_packages(exclude=["tests", "tests.*"]),
      install_requires=[
      ],
      extras_require={
          'test': [
              'pytest'
          ],
          'dev': [
              'ipython',
              'python-language-server[all]'
          ],
          'doc': [
              'sphinx',
              'sphinx_rtd_theme']})
